﻿using System;
using System.Linq;
using System.Reflection;
using System.Windows.Controls;
using System.Windows.Navigation;
using System.Collections.Generic;

using System.ComponentModel.Composition;
using System.ComponentModel.Composition.Hosting;

using Utils.Silverlight.MVVM.Navigation;

namespace Utils.Silverlight.MVVM
{
    [Export]
    public class CompositionNavigationContentLoader : INavigationContentLoader
    {
        /// <summary>
        /// A list of factories for all available Views.
        /// </summary>
        [ImportMany(typeof(IView))]
        public IEnumerable<ExportFactory<IView, IViewMetadata>> ViewExports { get; set; }

        /// <summary>
        /// A list of factories for all available ViewModels.
        /// </summary>
        [ImportMany(typeof(IViewModel))]
        public IEnumerable<ExportFactory<IViewModel, IViewModelMetadata>> ViewModelExports { get; set; }
  
        /// <summary>
        /// Load is enabled by default.
        /// </summary>
        /// <param name="targetUri"></param>
        /// <param name="currentUri"></param>
        /// <returns></returns>
        public bool CanLoad(Uri targetUri, Uri currentUri)
        {
            return true;
        }

        /// <summary>
        /// Don't do anything when canceling.
        /// </summary>
        /// <param name="asyncResult"></param>
        public void CancelLoad(IAsyncResult asyncResult)
        {
            return;
        }

        /// <summary>
        /// Begins asynchronous loading of the content for the specified target URI.
        /// </summary>
        /// <param name="targetUri">The URI to load content for.</param>
        /// <param name="currentUri">The URI that is currently loaded.</param>
        /// <param name="userCallback">The method to call when the content finishes loading.</param>
        /// <param name="asyncState">An object for storing custom state information.</param>
        /// <returns>
        /// An object that stores information about the asynchronous operation.
        /// </returns>
        public IAsyncResult BeginLoad(Uri targetUri, Uri currentUri, AsyncCallback userCallback, object asyncState)
        {
            // Convert to a dummy relative Uri so we can access the host.
            var relativeUri = new Uri("http://" + targetUri.OriginalString, UriKind.Absolute);

            // Get the factory for the ViewModel.
            var viewModelMapping = ViewModelExports.FirstOrDefault(o => o.Metadata.Key.Equals(relativeUri.Host, StringComparison.OrdinalIgnoreCase));
            if (viewModelMapping == null)
                throw new InvalidOperationException(
                    String.Format("Unable to navigate to: {0}. Could not locate the ViewModel.", targetUri.OriginalString));

            // Get the factory for the View.
            var viewMapping = ViewExports.FirstOrDefault(o => o.Metadata.ViewModelContract == viewModelMapping.Metadata.ViewModelContract);
            if (viewMapping == null)
                throw new InvalidOperationException(
                    String.Format("Unable to navigate to: {0}. Could not locate the View.", targetUri.OriginalString));

            // Resolve both the View and the ViewModel.
            var viewFactory = viewMapping.CreateExport();
            var view = viewFactory.Value as Control;
            var viewModelFactory = viewModelMapping.CreateExport();
            var viewModel = viewModelFactory.Value as IViewModel;

            // Attach ViewModel to View.
            view.DataContext = viewModel;
            viewModel.OnLoaded();

            // Get navigation values.
            var values = viewModelMapping.Metadata.GetArgumentValues(targetUri);
            viewModel.OnNavigated(values);

            if (view is Page)
            {
                Page page = view as Page;
                page.Title = viewModel.GetTitle();
            }
            else if (view is ChildWindow)
            {
                ChildWindow window = view as ChildWindow;
                window.Title = viewModel.GetTitle();
            }

            // Do not navigate if it's a ChildWindow.
            if (view is ChildWindow)
            {
                ProcessChildWindow(view as ChildWindow, viewModel);
                return null;
            }
            else
            {
                // Navigate because it's a Control.
                var result = new CompositionNavigationAsyncResult(asyncState, view);
                userCallback(result);
                return result;
            }
        }

        /// <summary>
        /// Show the ChildWindow and attach the CloseView event.
        /// </summary>
        /// <param name="window"></param>
        /// <param name="viewModel"></param>
        private void ProcessChildWindow(ChildWindow window, IViewModel viewModel)
        {
            // Close the ChildWindow if the ViewModel requests it.
            var closableViewModel = viewModel as IClosableViewModel;
            if (closableViewModel != null)
            {
                closableViewModel.CloseView += (s, e) => { window.Close(); };
            }

            // Show the window.
            window.Show();
        }

        /// <summary>
        /// Finished loading, proceed.
        /// </summary>
        /// <param name="asyncResult"></param>
        /// <returns></returns>
        public LoadResult EndLoad(IAsyncResult asyncResult)
        {
            return new LoadResult((asyncResult as CompositionNavigationAsyncResult).Result);
        }
    }
}
